/*
 * DefaultObject.java
 *
 * Created on 19. februar 2004, 18:40
 */


package Game;

import java.awt.Graphics2D;
import java.awt.Color;
import java.util.TreeSet;
import java.util.ArrayList;


/** A super class that all other game objects must inherit from */
public abstract class DefaultObject {
    
    
    class EventThread extends Thread {
        private EventThread() {
            super();
            setDaemon(true);
            setPriority(Thread.MIN_PRIORITY);
        }
        
        public void run() {
            processEvent();
        }
    }
    
    
    /** Class to hold the event information */
    class EventInformation {
        volatile int type = -1;
        volatile EventThread thread = null; // The thread which executes the events. It is volatile to prevent caching of this object, which could leed to problems when more threads try to access it
        volatile DefaultObject eventObject = null;
    }
    
    
    static Game game = null; // Holds a reference to the current game. Is private to prevent cheating
    
    // Event information
    EventInformation eventInformation = new EventInformation(); // Holds the information about the current event (if any)
    ArrayList dontThrowEventsToTheseBots = new ArrayList(); // Don't send events to these bots (If this is an explosion, it should only trigger the onHit event once)
    private volatile TreeSet ignoredEvents = new TreeSet(); // Insures no duplicates
    protected final static int onNoEvent = 0,
    onHit = 1,
    onDeath = 2,
    onCollision = 3,
    onSight = 4,
    onTargetHit = 5,
    onStart = 6;
    
    // Graphical representation
    double x = 1, y = 1, angle = 0; // x and y coordinates, angle in degrees
    double width = 30, height = 30;
    Color color = Color.BLACK;      // Color of object
    
    // Characteristics
    String onTeam = "";
    boolean isDead = false;      // If dead (ie. dying) the step() function should stop
    double firePower = 0;        // If object has fire power > 0 it will be destroyed upon collition, and the colliding object will loose this amount of health
    DefaultObject spawnedBy;     // The object, if any, that spawned this.
    DefaultObject rememberSpawnedBy; // The object, if any, that originally spawned this. If this is a bomb or rocket, the spawnedBy will eventually be set to null
    double healthStart = 100;      // Used when reseting the object
    double health = healthStart; // When reaching 0, the object (probably) dies
    boolean isGhost = false;     // Object does can move through other objects (no collition detection)
    
    
    /** Creates a new instance of DefaultObject */
    protected DefaultObject(double x, double y, double a, double w, double h, Color color) {
        this.x = x;
        this.y = y;
        this.angle = a;
        this.width = w;
        this.height = h;
        this.color = color;
    }
    
    
    /** Creates a new instance of DefaultObject. Used by Bot.java so the user does not have to call a complex constructor */
    DefaultObject() {
    }
    
    
    final void setGame(Game game) {
        if (this.game == null)
            this.game = game;
    }
    
    
    final void spawn(DefaultObject object) {
        game.spawn(object);
    }
    
    
    final void kill() {
        game.kill(this);
    }
    
    
    final void statsAddKill() {
        game.setup.statsAddKill((Bot) this);
    }
    
    
    final void statsAddDeath() {
        game.setup.statsAddDeath((Bot) this);
    }
    
    
    final void statsAddWin() {
        game.setup.statsAddWin((Bot) this);
    }
    
    
    final void statsAddPain(double pain) {
        game.setup.statsAddPain((Bot) this, pain);
    }
    
    
    final double getX() {
        return x;
    }
    
    
    final double getY() {
        return y;
    }
    
    
    final double getAngle() {
        return angle;
    }
    
    
    final double getWidth() {
        return width;
    }
    
    
    final double getHeight() {
        return height;
    }
    
    
    final Color getColor() {
        return color;
    }
    
    
    final boolean isDead() {
        return isDead;
    }
    
    
    final boolean isGhost() {
        return isGhost;
    }
    
    
    boolean isBullet() {
        return (this instanceof Bullet);
    }
    
    
    boolean isObstacle() {
        return (this instanceof Obstacle);
    }
    
    
    boolean isBot() {
        return (this instanceof Bot);
    }
    
    
    boolean isCustom() {
        return (getClass().getPackage() == null);
    }
    
    
    boolean isBomb() {
        return (this instanceof Bomb || this instanceof Rocket);
    }
    
    
    /** This is currently only used by Arena.java */
    final boolean isColliding() {
        
        // The text below culd be used as a reference for redesigning this function to test for collision (or containment) with 2 ellipses
        
/*
 *"Steven Jones" <_jones57@swbell.net> wrote in message
news:V4T_7.199$Cg7.75908236@newssvr12.news.prodigy.com...
> I need an algorithm to locate the points of intersection of two arbitrary
> 2D ellipses. I have solved the special case for ellipse which are not
> rotated but can not crack the more general nut.  For my application either
> ellipse may be rotated relative to the coordinate frame.
 
The following formal argument handles this.  Consider two
quadratic polynomials f(x) = a2*x^2+a1*x+a0 and
g(x) = b2*x^2+b1*x+b0.  If f(x) = 0 and g(x) = 0, then
  0 =  b2*f(x) - a2*g(x)
     = (a1*b2 - a2*b1)*x + (a0*b2 - a2*b0)
so x = (a2*b0 - a0*b2)/(a1*b2 - a2*b1)
Similarly,
  0 = b0*f(x) - a0*g(x)
     = (a2*b0-a0*b2)*x^2 + (a1*b0-a0*b1)*x
Divide by x to obtain
  0 = (a2*b0-a0*b2)*x + (a1*b0-a0*b1)
Now replace the earlier formula for x in this one to obtain
  0 = (a2*b0-a0*b2)^2/(a1*b2 - a2*b1) + (a1*b0-a0*b1)
Rearrange to get
  c = (a2*b1 - a1*b2)*(a1*b0 - a0*b1) - (a2*b0 - a0*b2)^2 = 0
This last condition is necessary for both f(x) = 0 and g(x) = 0.
 
Now the ellipse intersection problem.  One ellipse is specified
by F(x,y) = a0(y)+a1(y)*x+a2(y)*x^2 = 0 where a0(y) is
quadratic in y, a1(y) is linear in y, and a2(y) is a constant.
Similarly, G(x,y) = b0(y)+b1(y)*x_b2(y)*x^2 = 0 for the
other ellipse.  From the last paragraph, the expression c
depends on y.  Formally it is a 4th degree polynomial in y.
For each root y of c(y) = 0, use the equation that provides
x in terms of the 'a' and 'b' coefficients.
 
Numerically you need to worry about the leading coefficients
of c(y) being zero or nearly zero.  For example, if you really
have two circles, then c(y) is 2nd degree.
 
 */
        
        
        ArrayList objects = Game.getObjects();
        for (int i = 0; i < objects.size(); i++) { // Check for collision on all objects in the game
            DefaultObject object = (DefaultObject) objects.get(i);
            if (object != this // Don't collide with itself
            && (Math.abs(x - object.x) < width / 2 + object.width / 2)    // Collision on x-axis
            && (Math.abs(y - object.y) < height / 2 + object.height / 2)) // Collision on y-axis
                return true;
        }
        
        return false;
    }
    
    
    /** Tries to move the object to the new location by calculation collisions with other objects in the window */
    final void moveTo(double moveToX, double moveToY) {
        ArrayList objects = Game.getObjects();
        if (!isDead) {                                                         // If the object is dead, don't bother about collision detection
            for (int i = 0; i < objects.size(); i++) {                         // Check for collision on all objects in the game
                DefaultObject object = (DefaultObject) objects.get(i);
                
                // Check if collision when moveing to new coordinates
                if (object != this                                             // Don't collide with itself
                && object.spawnedBy != this                                    // Insures a bot does not shoot itself or collide with on bullets/bombs/rockets
                && spawnedBy != object                                         // Insures a bot does not shoot itself or collide with on bullets/bombs/rockets
                && !object.isDead                                              // If the object is Dead, don't bother about collision detection
                && (Math.abs(moveToX - object.x) < width / 2 + object.width / 2)      // Collision on x-axis
                && (Math.abs(moveToY - object.y) < height / 2 + object.height / 2)) { // Collision on y-axis
                    
                    // Check if collition when only moveing either the x or the y coordinate. If not, move only one. If collition with both, move none. This insures objects will not overlap
                    if (!(isGhost || object.isGhost)) {                        // If one of the objects is a ghost, don't try to withdraw movement
                        if (!((Math.abs(x - object.x) < width / 2 + object.width / 2) && (Math.abs(moveToY - object.y) < height / 2 + object.height / 2))) {
                            moveToX = x;
                        } else if (!((Math.abs(moveToX - object.x) < width / 2 + object.width / 2) && (Math.abs(y - object.y) < height / 2 + object.height / 2))) {
                            moveToY = y;
                        } else {
                            moveToX = x;
                            moveToY = y;
                        }
                    }
                    
                    // This is what happens on a collision
                    if (firePower > 0) {                                       // This is a bullet or an exploding bomb/rocket
                        if (Game.isDamageOnFriendlyFire || !onTeam.equals(object.onTeam)) { // Object is not a teammate
                            if (object.isBot()) {                              // Object is a bot
                                if (!dontThrowEventsToTheseBots.contains(object)) { // If a onHit event has already been thrown to the bot, don't do it again
                                    object.throwEventOnHit(this);              // Throw an event to the bot that was hit
                                    dontThrowEventsToTheseBots.add(object);    // Don't throw events to that bot from this object anymore. Does only have effect if this is an explosion
                                }
                                if (spawnedBy != null) {                       // This is not an exploding bomb or rocket
                                    spawnedBy.throwEventOnTargetHit(object);   // Throw an event to the creater of this bullet. Bombs and rockets will not throw events
                                    spawnedBy.statsAddPain(firePower);         // Add to the pain statistics for the owner of this bullet
                                } else if (rememberSpawnedBy != null) {        // This is an exploding bomb or rocket
                                    rememberSpawnedBy.throwEventOnTargetHit(object); // Throw an event to the creater of this bullet. Bombs and rockets will not throw events
                                    rememberSpawnedBy.statsAddPain(firePower); // Add to the pain statistics for the owner of this exploding bomb or rocket
                                }
                            }
                            object.looseHealth(firePower);                     // The object will loose an amount of health. This comes before the events are thrown to be sure the onDeath event will not be overridden with onHit
                            if (object.isBot() && object.isDead()) {           // If this was enought to kill this bot
                                object.throwEventOnDeath(this);
                                object.statsAddDeath();
                                if (spawnedBy != null)
                                    spawnedBy.statsAddKill();
                                else if (rememberSpawnedBy != null)
                                    rememberSpawnedBy.statsAddKill();
                            }
                        }
                        throwEventOnCollision(object);
                    } else if (object.firePower == 0) {                        // This is a bot colliding with something non-leathal
                        throwEventOnCollision(object);                         // Throw an event to this bot
                        object.throwEventOnCollision(this);                    // Throw an event to the object
                    }
                    
                    if (object.firePower > 0                                   // The object is a bullet or an exploding bomb/rocket
                    && !(isBot() && object.isBomb())                           // This is not a bot colliding with a bomb (it is calculated as a bomb colliding with a bot to prevent double dammage when moving)
                    && (Game.isDamageOnFriendlyFire || !object.onTeam.equals(onTeam))) { // Object was not fired from a teammate
                        if (isBot()) {                                         // The object is not a bot or obstacle
                            throwEventOnHit(object);                           // Throw an event to this one
                            object.throwEventOnCollision(this);
                            if (object.spawnedBy != null) {                    // Object is not an exploding bomb or rocket
                                object.spawnedBy.throwEventOnTargetHit(this);  // Throw an event to the creater of that bullet or bomb
                                object.spawnedBy.statsAddPain(object.firePower);
                            } else if (object.rememberSpawnedBy != null) {
                                object.rememberSpawnedBy.throwEventOnTargetHit(this); // Throw an event to the creater of that bullet or bomb
                                object.rememberSpawnedBy.statsAddPain(object.firePower);                                
                            }
                            looseHealth(object.firePower);                     // This object will loose an amount of health. This comes before the events are thrown to be sure the onDeath event will not be overridden with onHit
                            if (isDead()) {                                    // If this was enought to kill this bot
                                throwEventOnDeath(object);
                                statsAddDeath();
                                if (object.spawnedBy != null)
                                    object.spawnedBy.statsAddKill();
                                else if (object.rememberSpawnedBy != null)
                                    object.rememberSpawnedBy.statsAddKill();
                            }
                        }
                    }
                }
            }
        }
        x = moveToX;
        y = moveToY;
    }
    
    
    void looseHealth(double firePower) {
        health += - firePower;
        if (health <= 0) {
            die();
        }
    }
    
    
    /** Kill the object */
    void die() {
    }
    
    
    /** Draws the object on the screen  */
    abstract void draw(Graphics2D g2d);
    
    
    /** Handling of movement etc. */
    void step() {
    }
    
    
    /** Ignore an event */
    protected final void ignore(int ignoreEvent) {
        ignoredEvents.add(new Integer(ignoreEvent));
    }
    
    
    /** Ignore some events */
    protected final void ignore(int ignoreEvent1, int ignoreEvent2) {
        ignoredEvents.add(new Integer(ignoreEvent1));
        ignoredEvents.add(new Integer(ignoreEvent2));
    }
    
    
    /** Ignore some events */
    protected final void ignore(int ignoreEvent1, int ignoreEvent2, int ignoreEvent3) {
        ignoredEvents.add(new Integer(ignoreEvent1));
        ignoredEvents.add(new Integer(ignoreEvent2));
        ignoredEvents.add(new Integer(ignoreEvent3));
    }
    
    
    /** Ignore some events */
    protected final void ignore(int ignoreEvent1, int ignoreEvent2, int ignoreEvent3, int ignoreEvent4) {
        ignoredEvents.add(new Integer(ignoreEvent1));
        ignoredEvents.add(new Integer(ignoreEvent2));
        ignoredEvents.add(new Integer(ignoreEvent3));
        ignoredEvents.add(new Integer(ignoreEvent4));
    }
    
    
    /** Ignore some events */
    protected final void ignore(int ignoreEvent1, int ignoreEvent2, int ignoreEvent3, int ignoreEvent4, int ignoreEvent5) {
        ignoredEvents.add(new Integer(ignoreEvent1));
        ignoredEvents.add(new Integer(ignoreEvent2));
        ignoredEvents.add(new Integer(ignoreEvent3));
        ignoredEvents.add(new Integer(ignoreEvent4));
        ignoredEvents.add(new Integer(ignoreEvent5));
    }
    
    
    /** Ignore all event */
    protected final void ignore() {
        ignoredEvents.add(new Integer(onHit));
        ignoredEvents.add(new Integer(onDeath));
        ignoredEvents.add(new Integer(onCollision));
        ignoredEvents.add(new Integer(onSight));
        ignoredEvents.add(new Integer(onTargetHit));
        ignoredEvents.add(new Integer(onStart));
        ignoredEvents.add(new Integer(onNoEvent));
    }
    
    
    /** Unignore an event */
    protected final void unignore(int unignoreEvent) {
        ignoredEvents.remove(new Integer(unignoreEvent));
    }
    
    
    /** Unignore some events */
    protected final void unignore(int unignoreEvent1, int unignoreEvent2) {
        ignoredEvents.remove(new Integer(unignoreEvent1));
        ignoredEvents.remove(new Integer(unignoreEvent2));
    }
    
    
    /** Unignore some events */
    protected final void unignore(int unignoreEvent1, int unignoreEvent2, int unignoreEvent3) {
        ignoredEvents.remove(new Integer(unignoreEvent1));
        ignoredEvents.remove(new Integer(unignoreEvent2));
        ignoredEvents.remove(new Integer(unignoreEvent3));
    }
    
    
    /** Unignore some events */
    protected final void unignore(int unignoreEvent1, int unignoreEvent2, int unignoreEvent3, int unignoreEvent4) {
        ignoredEvents.remove(new Integer(unignoreEvent1));
        ignoredEvents.remove(new Integer(unignoreEvent2));
        ignoredEvents.remove(new Integer(unignoreEvent3));
        ignoredEvents.remove(new Integer(unignoreEvent4));
    }
    
    
    /** Unignore some events */
    protected final void unignore(int unignoreEvent1, int unignoreEvent2, int unignoreEvent3, int unignoreEvent4, int unignoreEvent5) {
        ignoredEvents.remove(new Integer(unignoreEvent1));
        ignoredEvents.remove(new Integer(unignoreEvent2));
        ignoredEvents.remove(new Integer(unignoreEvent3));
        ignoredEvents.remove(new Integer(unignoreEvent4));
        ignoredEvents.remove(new Integer(unignoreEvent5));
    }
    
    
    /** Unignore all event */
    protected final void unignore() {
        ignoredEvents.remove(new Integer(onHit));
        ignoredEvents.remove(new Integer(onDeath));
        ignoredEvents.remove(new Integer(onCollision));
        ignoredEvents.remove(new Integer(onSight));
        ignoredEvents.remove(new Integer(onTargetHit));
        ignoredEvents.remove(new Integer(onStart));
        ignoredEvents.remove(new Integer(onNoEvent));
    }
    
    
    /** Check if an event is in the ignore list */
    protected final boolean isIgnored(int event) {
        return ignoredEvents.contains(new Integer(event));
    }
    

    /** Create a new thread which executes the current event */
    private synchronized void makeNewThread(DefaultObject object) {
        stopThread();
        eventInformation.eventObject = object;
        eventInformation.thread = new EventThread();
        //eventInformation.thread.setPriority(Thread.MIN_PRIORITY);
        eventInformation.thread.start();
    }
    
    
    synchronized final void stopThread() {
        if (eventInformation.thread != null) {
            eventInformation.thread.stop(); // Depricated, but necessary! The alternative as proposed by Sun (http://java.sun.com/j2se/1.5.0/docs/guide/misc/threadPrimitiveDeprecation.html) is not possible since it is executing user-created code. This is the only way of stopping a thread no matter where in the code it is executing. This is not an issue since the object will (hopefully) not enter an unstable state if the monitor is unlocked suddenly
            eventInformation.thread = null;
        }
    }
    
    
    /** Set the event flag, stop the current thread and create a new thread that will execute the specified event */
    final synchronized void throwEventOnStart() {
        if (!isIgnored(onStart)) {
            eventInformation.type = onStart;
            makeNewThread(this);
        }
    }
    
    
    /** Set the event flag, stop the current thread and create a new thread that will execute the specified event */
    final synchronized void throwEventOnCollision(DefaultObject collidedWith) {
        if (!isIgnored(onCollision)) {
            eventInformation.type = onCollision;
            makeNewThread(collidedWith);
        }
    }
    
    
    /** Set the event flag, stop the current thread and create a new thread that will execute the specified event */
    final synchronized void throwEventOnSight(DefaultObject objectInSight) {
        if (!isIgnored(onSight)) {
            eventInformation.type = onSight;
            makeNewThread(objectInSight);
        }
    }
    
    
    /** Set the event flag, stop the current thread and create a new thread that will execute the specified event */
    final synchronized void throwEventOnHit(DefaultObject hitBy) {
        if (!isIgnored(onHit)) {
            eventInformation.type = onHit;
            makeNewThread(hitBy);
        }
    }
    
    
    /** Set the event flag, stop the current thread and create a new thread that will execute the specified event */
    final synchronized void throwEventOnTargetHit(DefaultObject targetHit) {
        if (!isIgnored(onTargetHit)) {
            eventInformation.type = onTargetHit;
            makeNewThread(targetHit);
        }
    }
    
    
    /** Set the event flag, stop the current thread and create a new thread that will execute the specified event */
    final synchronized void throwEventOnDeath(DefaultObject killedBy) {
        if (!isIgnored(onDeath)) {
            eventInformation.type = onDeath;
            makeNewThread(killedBy);
        }
    }
    
    
    /** Default event handler. If not overridden, ignore them in the future. Must be AT LEAST protected, or the event listener will not call the custom overridden methods */
    protected void onStart() {
        ignore(onStart);
    }
    
    
    /** Default event handler. If not overridden, ignore them in the future. Must be AT LEAST protected, or the event listener will not call the custom overridden methods */
    protected void onHit(double hitFromMovementAngle) {
        ignore(onHit);
    }
    
    
    /** Default event handler. If not overridden, ignore them in the future. Must be AT LEAST protected, or the event listener will not call the custom overridden methods */
    protected void onDeath(String killedBy) {
        ignore(onDeath); // Not too late, since some objects (like obstacles) don't care how many times they "die"
    }
    
    
    /** Default event handler. If not overridden, ignore them in the future. Must be AT LEAST protected, or the event listener will not call the custom overridden methods */
    protected void onCollision() {
        ignore(onCollision);
    }
    
    
    /** Default event handler. If not overridden, ignore them in the future. Must be AT LEAST protected, or the event listener will not call the custom overridden methods */
    protected void onSight(String type, double distance) {
        ignore(onSight);
    }
    
    
    /** Default event handler. If not overridden, ignore them in the future. Must be AT LEAST protected, or the event listener will not call the custom overridden methods */
    protected void onTargetHit() {
        ignore(onTargetHit);
    }
    
    
    /** Default event handler. If not overridden, ignore them in the future. Must be AT LEAST protected, or the event listener will not call the custom overridden methods */
    protected void onNoEvent() {
        ignore(onNoEvent);
    }
    
    
    /** This is the "home" of the current event thread. Must NOT be called from anything but the constructor of the current thread */
    private void processEvent() {
        while (eventInformation.thread == Thread.currentThread()) { // Make sure this thread will eventually exit
            try {
                switch (eventInformation.type) {
                    
                    case onNoEvent:
                        onNoEvent();
                        break;
                        
                    case onHit:
                        double absoluteAngle = Math.atan2(eventInformation.eventObject.getY() - y, eventInformation.eventObject.getX() - x);
                        double relativeAngle = absoluteAngle - angle;
                        double outputAngle = Math.toDegrees(relativeAngle);
                        if (outputAngle > 180)
                            outputAngle += -360;
                        else if (outputAngle < - 180)
                            outputAngle += 360;
                        
                        onHit(outputAngle);
                        break;
                        
                    case onDeath:
                        onDeath(eventInformation.eventObject.toString());
                        break;
                        
                    case onCollision:
                        onCollision();
                        break;
                        
                    case onSight:
                        String type = "BUG!";
                        if (eventInformation.eventObject.isBot() && Game.isTeamPlayOn && eventInformation.eventObject.onTeam == onTeam)
                            type = "teammate";
                        else if (eventInformation.eventObject.isBot() && (!Game.isTeamPlayOn || eventInformation.eventObject.onTeam != onTeam))
                            type = "enemy";
                        else if (eventInformation.eventObject.isObstacle())
                            type = "obstacle";
                        else if (eventInformation.eventObject.isBomb())
                            type = "bomb";
                        
                        if (type == "BUG!")
                            Setup.notice("BUG! In DefaultObject.processEvent(): Unknown type");
                        
                        double distance = Math.sqrt((x - eventInformation.eventObject.getX()) * (x - eventInformation.eventObject.getX()) + (y - eventInformation.eventObject.getY()) * (y - eventInformation.eventObject.getY()));
                        onSight(type, distance);
                        break;
                        
                    case onTargetHit:
                        onTargetHit();
                        break;
                        
                    case onStart:
                        onStart();
                        break;
                        
                    default:
                        Setup.notice("BUG! In DefaultObject.processEvent(): Unknown event: " + eventInformation.type);
                        
                }
                
            } catch (Exception e) {
                Game.spam(getClass().getName() + "has been disqualified!");
                die();
                System.out.println("\n\n" + getClass().getName() + " fucked it up:\n" + e.getClass().getName() + ":\n" + e.getMessage() + "\n");
            }
            
            if (isIgnored(onNoEvent)) {
                break; // Break the while loop and exit this thread (event)
            } else {   // Else re-run while command to access the onNoEvent method

                eventInformation.type = onNoEvent;
                try {
                    Thread.sleep(Game.aiSpeed);
                } catch (InterruptedException e) { break; }
                
            }
        }
    }
    
}
